==> newsstar-1.5.3.tar.gz: newsstar-1.5.3/src/kiddy.c <==
/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>

#include "child.h"
#include "deblog.h"
#include "intercom.h"
#include "kiddy.h"

static KiddyData *kiddy_table;

int kiddy_nthreads = 0;

/* We may have some extraneous child processes, because some may be disused
 * after reading server configs. maincf_get_max_threads() will be updated by
 * this. kiddy_total_threads is total number of threads, active or not */
static int kiddy_total_threads;

/* Parent reads commands from all children from this */
static Intercom *kiddy_intercom = NULL;

KiddyData *kiddy_get_available(void)
{
	int i;

	for (i = 0; i < maincf_get_max_threads(); ++i)
	{
		if (!kiddy_table[i].thread)
		{
			return &kiddy_table[i];
		}
	}
	return NULL;
}

void kiddy_detach_thread(KiddyData *kiddy)
{
	kiddy->thread = NULL;
}

#if !USE_PTHREADS
static void kiddy_sigchld_handler(int signum)
{
	int i;
	BOOL found_kid = FALSE;
	
	UNUSED(signum);

	for (i = 0; i < kiddy_total_threads; ++i)
	{
		if (kiddy_table[i].pid != MAIN_PID)
		{
			if (waitpid(kiddy_table[i].pid, NULL, WNOHANG))
			{
				deblog_critical(_("A child process has died unexpectedly\n"));
				found_kid = TRUE;
				kiddy_table[i].pid = MAIN_PID;
			}
		}
	}
	if (!found_kid && maincf_get_using_sn() &&
			!strcmp(maincf_get_post_method(), "snstore"))
	{
		deblog_error(_("snstore process appears to have died unexpectedly\n"));
	}
	if (!--kiddy_nthreads)
	{
		deblog_error(_("No child processes left, exiting\n"));
	}
}
#endif /* !USE_PTHREADS */

#if USE_PTHREADS
typedef struct {
	int thread_idx;
	IntercomQ *read_from;
	IntercomQ *write_to;
} KiddyThreadStartData;

static void *kiddy_start_pthread(void *data)
{
	KiddyThreadStartData *td = data;
	
	deblog_debug("pthreads starting thread %d\n", td->thread_idx);
	child_start(td->thread_idx, td->read_from, td->write_to);
	free(td);
	return NULL;
}
#endif

static pid_t kiddy_spawn(int thread_idx, int read_from, int write_to, int
		*pipe_table)
{
	kiddy_thread_t pid;
	
#if USE_PTHREADS
	KiddyThreadStartData *td = XMALLOC(sizeof(KiddyThreadStartData));
	
	td->thread_idx = thread_idx;
	td->read_from = intercom_queue_new(0, FALSE);
	deblog_debug("IQ: Thread %d will read from q %p\n",
			thread_idx, td->read_from);
	td->write_to = kiddy_intercom->queue;
	kiddy_table[thread_idx].queue = td->read_from;
	deblog_debug("kiddy_spawn creating thread %d\n", thread_idx);
	pthread_create(&pid, NULL, kiddy_start_pthread, td);
#else /* !USE_PTHREADS */
	pid = fork();
	switch (pid)
	{
		int i;
		
		case 0:
			/* Close all pipes we're not using */
			for (i = 0; i < kiddy_total_threads; ++i)
			{
				if (pipe_table[i] != -1)
				{
					close(pipe_table[i]);
					pipe_table[i] = -1;
				}
			}
			child_start(thread_idx, read_from, write_to);
			break;
		case -1:
		   deblog_error(_("Unable to fork: %s\n"), strerror(errno));
			break;
		default:
			/* Don't need to read from parent's copy of read_from, close it */
			close(read_from);
			return pid;
	}
#endif /* !USE_PTHREADS */
	return 0;
}

void kiddy_spawn_all()
{
	int i;
	int write_to[2];
	/* Need a table of all parent > child pipes so that children can close ones
	 * they aren't using */
	int *pipe_table;
	int flags;
#if USE_PTHREADS
	IntercomQ *q;
#endif
	
	kiddy_total_threads = maincf_get_max_threads();
	deblog_info(_("Starting %d child processes\n"), kiddy_total_threads);
	kiddy_table = XMALLOC(kiddy_total_threads * sizeof(KiddyData));
	pipe_table = XMALLOC(kiddy_total_threads * sizeof(int));
	for (i = 0; i < kiddy_total_threads; ++i)
	{
		kiddy_table[i].thread = NULL;
		kiddy_table[i].queue = NULL;
		kiddy_table[i].pid = MAIN_PID;
		kiddy_table[i].thread_idx = i;
		pipe_table[i] = -1;
	}
#if USE_PTHREADS
	/* Create before spawning child threads; they'll use it */
	q = intercom_queue_new(0, TRUE);
	deblog_debug("IQ: Created q %p for receiving command from children\n", q);
	kiddy_intercom = intercom_new(q);
#else
	signal(SIGCHLD, kiddy_sigchld_handler);
#endif
	XPIPE(write_to);
	flags = fcntl(write_to[0], F_GETFL, 0);
	if (fcntl(write_to[0], F_SETFL, flags | O_NONBLOCK) == -1)
	{
		deblog_critical(_("Can't make intercom pipe non-blocking\n"));
	}
	for (i = 0; i < kiddy_total_threads; ++i)
	{
		int pipefd[2] = {0, 0};

#if !USE_PTHREADS
		XPIPE(pipefd);
		deblog_debug(_("Child %d will read from %d and be written to via %d\n"),
				i, pipefd[0], pipefd[1]);
		pipe_table[i] = pipefd[1];
#endif
		kiddy_table[i].pid = kiddy_spawn(i, pipefd[0], write_to[1], pipe_table);
	}
#if !USE_PTHREADS
	/* Parent needs own IntercomQ for sending to each child; in pthread
	 * mode use the one created by kiddy_spawn. */
	for (i = 0; i < maincf_get_max_threads(); ++i)
	{
		kiddy_table[i].queue = intercom_queue_new(pipe_table[i], FALSE);
	}
	/* We don't need to write to write_to */
	close(write_to[1]);
	/* Create last to avoid duplicating in children */
	kiddy_intercom = intercom_new(intercom_queue_new(write_to[0], TRUE));
#endif
}

const intercom_data *kiddy_get_command(void)
{
	return intercom_get_command(kiddy_intercom);
}

NSThread *kiddy_get_next_thread(const NSThread *thread)
{
	int i = 0;
	BOOL found = (thread == NULL);

	deblog_info("kiddy_get_next_thread starting from %p\n", thread);
	if (!found)
	{
		for (i = 0; i < maincf_get_max_threads(); ++i)
		{
			deblog_info("Looking at thread %d/%p/%ld for starter\n",
					i, kiddy_table[i].thread, kiddy_table[i].pid);
			if (kiddy_table[i].pid != MAIN_PID &&
							kiddy_table[i].thread == thread)
			{
				deblog_info("It's a match\n");
				++i;
				break;
			}
		}
	}
	for (; i < maincf_get_max_threads(); ++i)
	{
		deblog_info("Looking at thread %d/%p/%ld for next\n",
				i, kiddy_table[i].thread, kiddy_table[i].pid);
		if (kiddy_table[i].thread && kiddy_table[i].pid != MAIN_PID)
		{
			deblog_info("It's a goer\n");
			return kiddy_table[i].thread;
		}
	}
	return NULL;
}

==> newsstar-1.5.6.tar.gz: newsstar-1.5.6/src/kiddy.c <==
/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>

#include "child.h"
#include "deblog.h"
#include "intercom.h"
#include "kiddy.h"

static KiddyData *kiddy_table;

int kiddy_nthreads = 0;

/* We may have some extraneous child processes, because some may be disused
 * after reading server configs. maincf_get_max_threads() will be updated by
 * this. kiddy_total_threads is total number of threads, active or not */
static int kiddy_total_threads;

/* Parent reads commands from all children from this */
static Intercom *kiddy_intercom = NULL;

KiddyData *kiddy_get_available(void)
{
	int i;

	for (i = 0; i < maincf_get_max_threads(); ++i)
	{
		if (!kiddy_table[i].thread)
		{
			return &kiddy_table[i];
		}
	}
	return NULL;
}

void kiddy_detach_thread(KiddyData *kiddy)
{
	kiddy->thread = NULL;
}

#if !USE_PTHREADS
static void kiddy_sigchld_handler(int signum)
{
	int i;
	BOOL found_kid = FALSE;

	UNUSED(signum);

	for (i = 0; i < kiddy_total_threads; ++i)
	{
		if (kiddy_table[i].pid != MAIN_PID)
		{
			if (waitpid(kiddy_table[i].pid, NULL, WNOHANG))
			{
				deblog_critical(_("A child process has died unexpectedly\n"));
				found_kid = TRUE;
				kiddy_table[i].pid = MAIN_PID;
			}
		}
	}
	if (!found_kid && maincf_get_using_sn() &&
			!strcmp(maincf_get_post_method(), "snstore"))
	{
		deblog_error(_("snstore process appears to have died unexpectedly\n"));
	}
	if (!--kiddy_nthreads)
	{
		deblog_error(_("No child processes left, exiting\n"));
	}
}
#endif /* !USE_PTHREADS */

#if USE_PTHREADS
typedef struct {
	int thread_idx;
	IntercomQ *read_from;
	IntercomQ *write_to;
} KiddyThreadStartData;

static void *kiddy_start_pthread(void *data)
{
	KiddyThreadStartData *td = data;

	deblog_debug("pthreads starting thread %d\n", td->thread_idx);
	child_start(td->thread_idx, td->read_from, td->write_to);
	free(td);
	return NULL;
}
#endif

static pid_t kiddy_spawn(int thread_idx, int read_from, int write_to, int
		*pipe_table)
{
	kiddy_thread_t pid;

#if USE_PTHREADS
	KiddyThreadStartData *td = XMALLOC(sizeof(KiddyThreadStartData));

	td->thread_idx = thread_idx;
	td->read_from = intercom_queue_new(0, FALSE);
	deblog_debug("IQ: Thread %d will read from q %p\n",
			thread_idx, td->read_from);
	td->write_to = kiddy_intercom->queue;
	kiddy_table[thread_idx].queue = td->read_from;
	deblog_debug("kiddy_spawn creating thread %d\n", thread_idx);
	pthread_create(&pid, NULL, kiddy_start_pthread, td);
#else /* !USE_PTHREADS */
	pid = fork();
	switch (pid)
	{
		int i;

		case 0:
			/* Close all pipes we're not using */
			for (i = 0; i < kiddy_total_threads; ++i)
			{
				if (pipe_table[i] != -1)
				{
					close(pipe_table[i]);
					pipe_table[i] = -1;
				}
			}
			child_start(thread_idx,
			        rw_create_buf(read_from, FALSE, TRUE, -1),
			        rw_create_buf(write_to, FALSE, TRUE, -1));
			break;
		case -1:
		   deblog_error(_("Unable to fork: %s\n"), strerror(errno));
			break;
		default:
			/* Don't need to read from parent's copy of read_from, close it */
			close(read_from);
			return pid;
	}
#endif /* !USE_PTHREADS */
	return 0;
}

void kiddy_spawn_all()
{
	int i;
	int write_to[2];
	/* Need a table of all parent > child pipes so that children can close ones
	 * they aren't using */
	int *pipe_table;
	int flags;
#if USE_PTHREADS
	IntercomQ *q;
#endif

	kiddy_total_threads = maincf_get_max_threads();
	deblog_info(_("Starting %d child processes\n"), kiddy_total_threads);
	kiddy_table = XMALLOC(kiddy_total_threads * sizeof(KiddyData));
	pipe_table = XMALLOC(kiddy_total_threads * sizeof(int));
	for (i = 0; i < kiddy_total_threads; ++i)
	{
		kiddy_table[i].thread = NULL;
		kiddy_table[i].queue = NULL;
		kiddy_table[i].pid = MAIN_PID;
		kiddy_table[i].thread_idx = i;
		pipe_table[i] = -1;
	}
#if USE_PTHREADS
	/* Create before spawning child threads; they'll use it */
	q = intercom_queue_new(0, TRUE);
	deblog_debug("IQ: Created q %p for receiving command from children\n", q);
	kiddy_intercom = intercom_new(q);
#else
	signal(SIGCHLD, kiddy_sigchld_handler);
#endif
	XPIPE(write_to);
	flags = fcntl(write_to[0], F_GETFL, 0);
	if (fcntl(write_to[0], F_SETFL, flags | O_NONBLOCK) == -1)
	{
		deblog_critical(_("Can't make intercom pipe non-blocking\n"));
	}
	for (i = 0; i < kiddy_total_threads; ++i)
	{
		int pipefd[2] = {0, 0};

#if !USE_PTHREADS
		XPIPE(pipefd);
		deblog_debug(_("Child %d will read from %d and be written to via %d\n"),
				i, pipefd[0], pipefd[1]);
		pipe_table[i] = pipefd[1];
#endif
		kiddy_table[i].pid = kiddy_spawn(i, pipefd[0], write_to[1], pipe_table);
	}
#if !USE_PTHREADS
	/* Parent needs own IntercomQ for sending to each child; in pthread
	 * mode use the one created by kiddy_spawn. */
	for (i = 0; i < maincf_get_max_threads(); ++i)
	{
		kiddy_table[i].queue = intercom_queue_new(pipe_table[i], FALSE);
	}
	/* We don't need to write to write_to */
	close(write_to[1]);
	/* Create last to avoid duplicating in children */
	kiddy_intercom = intercom_new(intercom_queue_new(write_to[0], TRUE));
#endif
}

const intercom_data *kiddy_get_command(void)
{
	return intercom_get_command(kiddy_intercom);
}

NSThread *kiddy_get_next_thread(const NSThread *thread)
{
	int i = 0;
	BOOL found = (thread == NULL);

	deblog_info("kiddy_get_next_thread starting from %p\n", thread);
	if (!found)
	{
		for (i = 0; i < maincf_get_max_threads(); ++i)
		{
			deblog_info("Looking at thread %d/%p/%ld for starter\n",
					i, kiddy_table[i].thread, kiddy_table[i].pid);
			if (kiddy_table[i].pid != MAIN_PID &&
							kiddy_table[i].thread == thread)
			{
				deblog_info("It's a match\n");
				++i;
				break;
			}
		}
	}
	for (; i < maincf_get_max_threads(); ++i)
	{
		deblog_info("Looking at thread %d/%p/%ld for next\n",
				i, kiddy_table[i].thread, kiddy_table[i].pid);
		if (kiddy_table[i].thread && kiddy_table[i].pid != MAIN_PID)
		{
			deblog_info("It's a goer\n");
			return kiddy_table[i].thread;
		}
	}
	return NULL;
}
